==> fireducks-0.10.3-cp311-cp311-manylinux_2_28_x86_64.whl: fireducks/pandas/indexing.py <==
# Copyright (c) 2023 NEC Corporation. All Rights Reserved.

import logging

import fireducks.core
import fireducks.pandas.utils as utils
from fireducks import ir, irutils

logger = logging.getLogger(__name__)


class _Indexer:
    def __init__(self, obj, name):
        self.obj = obj
        self.name = name

    def _unwrap(self, reason=None):
        return getattr(self.obj._unwrap(reason=reason), self.name)

    def __getitem__(self, key):
        reason = "fallback _Indexer(" + self.name + ").__getitem__"
        return utils.fallback_call(
            self._unwrap, "__getitem__", key, __fireducks_reason=reason
        )

    def __setitem__(self, key, value):
        reason = "fallback _Indexer(" + self.name + ").__setitem__"
        utils.fallback_call(
            self._unwrap, "__setitem__", key, value, __fireducks_reason=reason
        )
        self.obj._rebind_to_cache()


def is_column_locator(obj):
    if isinstance(obj, str):
        return True
    if irutils._is_str_list(obj):
        return True
    return False


def is_full_slice(slobj):
    return slobj.start is None and slobj.stop is None and slobj.step is None


class _LocIndexer(_Indexer):
    def __init__(self, obj):
        super().__init__(obj, "loc")

    def __getitem__(self, key):
        from fireducks.pandas.series import Series

        t_cls = self.obj.__class__

        if isinstance(key, Series):
            dtype = utils._deduce_dtype(key)
            if dtype is not None and dtype == bool:
                return t_cls._create(ir.filter(self.obj._value, key._value))

        if isinstance(key, tuple) and len(key) == 2:
            r, c = key
            # get column
            if isinstance(r, slice) and is_column_locator(c):
                if is_full_slice(r):  # e.g., loc[:, ['a', 'b']]
                    return self.obj[c]

            if isinstance(r, Series):
                supported_c = True
                if is_column_locator(c):
                    projected = self.obj[c]
                elif isinstance(c, slice) and is_full_slice(c):
                    # TODO: support column slicing
                    projected = self.obj
                else:
                    supported_c = False

                if supported_c:
                    dtype = utils._deduce_dtype(r)
                    if dtype is not None and dtype == bool:
                        return projected.__class__._create(
                            ir.filter(projected._value, r._value)
                        )
                    else:  # TODO: support index selection
                        return _LocIndexer(projected)[r]

            if isinstance(r, slice) and isinstance(c, slice):
                if is_full_slice(r) and is_full_slice(c):  # e.g., loc[:, :]
                    return self.obj

            # get single element
            if isinstance(r, int) and isinstance(c, str):
                if (
                    fireducks.core.get_fireducks_options().fast_fallback
                    and self.obj._fireducks_meta.is_cached()
                ):
                    cache = self.obj._fireducks_meta.get_cache()
                    return cache.loc[key]

        reason = "fallback _LocIndexer(" + self.name + ").__getitem__"
        return utils.fallback_call(
            self._unwrap, "__getitem__", key, __fireducks_reason=reason
        )

    def __setitem__(self, key, val):
        from fireducks.pandas import Series

        checked = False
        if isinstance(self.obj, Series):
            if (
                isinstance(key, Series) or irutils._is_index_like(key)
            ) and irutils.irable_scalar(val):
                checked = True
                from_non_indexed_arraylike = not isinstance(key, Series)
                index_or_mask = (
                    Series(key) if from_non_indexed_arraylike else key
                )
                value = irutils.make_scalar(val)
                cname = irutils.make_column_name(None)
                is_series = True
        else:  # DataFrame
            if isinstance(key, tuple) and len(key) == 2:
                r, c = key
                if (
                    (isinstance(r, Series) or irutils._is_index_like(r))
                    and isinstance(c, str)
                    and irutils.irable_scalar(val)  # TODO: support other cases
                ):
                    checked = True
                    from_non_indexed_arraylike = not isinstance(r, Series)
                    index_or_mask = (
                        Series(r) if from_non_indexed_arraylike else r
                    )
                    value = irutils.make_scalar(val)
                    cname = irutils.make_column_name(c)
                    is_series = False

        if checked:
            res_value = ir.loc_setter_with_scalar(
                self.obj._value,
                index_or_mask._value,
                value,
                cname,
                from_non_indexed_arraylike,
                is_series,
            )
            return self.obj._rebind(res_value, invalidate_cache=True)

        reason = "fallback _LocIndexer(" + self.name + ").__setitem__"
        utils.fallback_call(
            self._unwrap, "__setitem__", key, val, __fireducks_reason=reason
        )
        return self.obj._rebind_to_cache()


class _IlocIndexer(_Indexer):
    def __init__(self, obj):
        super().__init__(obj, "iloc")

    def __getitem__(self, key):
        from fireducks.pandas.series import Series
        from fireducks.pandas.generic import _Scalar

        if isinstance(self.obj, Series) and isinstance(key, int):
            # get single element
            if (
                fireducks.core.get_fireducks_options().fast_fallback
                and self.obj._fireducks_meta.is_cached()
            ):  # read from cache
                cache = self.obj._fireducks_meta.get_cache()
                return cache.iloc[key]
            else:
                value = ir.iloc_scalar(self.obj._value, key)
                return _Scalar(value)._unwrap()

        # self.obj must be DataFrame
        if isinstance(key, tuple) and len(key) == 2:
            r, c = key
            # get single element
            if isinstance(r, int) and isinstance(c, int):
                if (
                    fireducks.core.get_fireducks_options().fast_fallback
                    and self.obj._fireducks_meta.is_cached()
                ):  # read from cache
                    cache = self.obj._fireducks_meta.get_cache()
                    return cache.iloc[key]
                else:
                    columns = self.obj.columns
                    projected = self.obj[columns[c]]
                    value = ir.iloc_scalar(projected._value, r)
                    return _Scalar(value)._unwrap()

            # projection
            if isinstance(r, slice) and isinstance(c, slice):
                if is_full_slice(r):
                    # e.g., iloc[:, :]
                    if is_full_slice(c):
                        return self.obj
                    # e.g., iloc[:, :2]
                    else:
                        columns = self.obj.columns
                        return self.obj[columns[c]]

        reason = "fallback _ILocIndexer(" + self.name + ").__getitem__"
        return utils.fallback_call(
            self._unwrap, "__getitem__", key, __fireducks_reason=reason
        )

==> fireducks-0.12.4-cp311-cp311-manylinux_2_17_x86_64.whl: fireducks/pandas/indexing.py <==
# Copyright (c) 2023 NEC Corporation. All Rights Reserved.

import logging
import numpy as np

import fireducks.core
import fireducks.pandas.utils as utils
from fireducks import ir, irutils

logger = logging.getLogger(__name__)


class _Indexer:
    def __init__(self, obj, name):
        self.obj = obj
        self.name = name

    def _unwrap(self, reason=None):
        return getattr(self.obj._unwrap(reason=reason), self.name)

    def __getitem__(self, key):
        reason = "fallback _Indexer(" + self.name + ").__getitem__"
        return utils.fallback_call(
            self._unwrap, "__getitem__", key, __fireducks_reason=reason
        )

    def __setitem__(self, key, value):
        reason = "fallback _Indexer(" + self.name + ").__setitem__"
        self.obj._fallback_mutating_method(
            self.name + ".__setitem__",
            args=[key, value],
            reason=reason,
        )


def is_column_locator(obj):
    if isinstance(obj, str):
        return True
    if irutils._is_str_list(obj):
        return True
    return False


def is_full_slice(slobj):
    return (
        isinstance(slobj, slice)
        and slobj.start is None
        and slobj.stop is None
        and slobj.step is None
    )


class _LocIndexer(_Indexer):
    def __init__(self, obj):
        super().__init__(obj, "loc")

    def __getitem__(self, key):
        from fireducks.pandas.series import Series

        t_cls = self.obj.__class__

        if isinstance(key, Series):
            dtype = utils._deduce_dtype(key)
            if dtype is not None and dtype == bool:
                return t_cls._create(ir.filter(self.obj._value, key._value))

        if isinstance(key, tuple) and len(key) == 2:
            r, c = key
            # get column: e.g., loc[:, ['a', 'b']]
            if is_full_slice(r) and is_column_locator(c):
                return self.obj[c]

            if isinstance(r, Series):
                supported_c = True
                if is_column_locator(c):
                    projected = self.obj[c]
                elif is_full_slice(c):
                    # TODO: support column slicing
                    projected = self.obj
                else:
                    supported_c = False

                if supported_c:
                    dtype = utils._deduce_dtype(r)
                    if dtype is not None and dtype == bool:
                        return projected.__class__._create(
                            ir.filter(projected._value, r._value)
                        )
                    else:  # TODO: support index selection
                        return _LocIndexer(projected)[r]

            if is_full_slice(r) and is_full_slice(c):  # e.g., loc[:, :]
                return self.obj

            # get single element
            if isinstance(r, int) and isinstance(c, str):
                if (
                    fireducks.core.get_fireducks_options().fast_fallback
                    and self.obj._fireducks_meta.is_cached()
                ):
                    cache = self.obj._fireducks_meta.get_cache()
                    return cache.loc[key]

        reason = "fallback _LocIndexer(" + self.name + ").__getitem__"
        return utils.fallback_call(
            self._unwrap, "__getitem__", key, __fireducks_reason=reason
        )

    def __setitem__(self, key, val):
        from fireducks.pandas import Series

        checked = False
        if isinstance(self.obj, Series):
            if (
                isinstance(key, Series) or irutils._is_index_like(key)
            ) and irutils.irable_scalar(val):
                checked = True
                from_non_indexed_arraylike = not isinstance(key, Series)
                index_or_mask = (
                    Series(key) if from_non_indexed_arraylike else key
                )
                value = irutils.make_scalar(val)
                cname = irutils.make_column_name(None)
                is_series = True
        else:  # DataFrame
            if isinstance(key, tuple) and len(key) == 2:
                r, c = key
                if (
                    (isinstance(r, Series) or irutils._is_index_like(r))
                    and isinstance(c, str)
                    and irutils.irable_scalar(val)  # TODO: support other cases
                ):
                    checked = True
                    from_non_indexed_arraylike = not isinstance(r, Series)
                    index_or_mask = (
                        Series(r) if from_non_indexed_arraylike else r
                    )
                    value = irutils.make_scalar(val)
                    cname = irutils.make_column_name(c)
                    is_series = False

        if checked:
            res_value = ir.loc_setter_with_scalar(
                self.obj._value,
                index_or_mask._value,
                value,
                cname,
                from_non_indexed_arraylike,
                is_series,
            )
            return self.obj._rebind(res_value, invalidate_cache=True)

        reason = "fallback _LocIndexer(" + self.name + ").__setitem__"
        self.obj._fallback_mutating_method(
            self.name + ".__setitem__",
            args=[key, val],
            reason=reason,
        )


class _IlocIndexer(_Indexer):
    def __init__(self, obj):
        super().__init__(obj, "iloc")

    def __getitem__(self, key):
        from fireducks.pandas import Series, Index
        from fireducks.pandas.generic import _Scalar

        # slice rows of input frame/series: iloc[range(2)]
        if isinstance(key, range):
            key = slice(key.start, key.stop, key.step)

        # slice rows of input frame/series: iloc[:2]
        if isinstance(key, slice):
            return self.obj._slice(key)

        # filter rows of input frame/series based on given indices...
        if isinstance(key, (list, Series, Index, np.ndarray)):  # iloc[[0,3]]
            return self.obj._take(key, axis=0)

        def _take_col(col_idx):
            # take doesn't support int
            idx = irutils.make_vector_or_scalar_of_scalar(col_idx)
            return Series._create(
                ir.take(
                    self.obj._value,
                    idx,
                    axis=1,
                    ignore_index=False,
                )
            )

        if isinstance(self.obj, Series):
            if isinstance(key, int):
                # get single element
                if (
                    fireducks.core.get_fireducks_options().fast_fallback
                    and self.obj._fireducks_meta.is_cached()
                ):  # read from cache
                    cache = self.obj._fireducks_meta.get_cache()
                    return cache.iloc[key]
                else:
                    value = ir.iloc_scalar(self.obj._value, key)
                    return _Scalar(value)._unwrap()

            reason = "fallback _ILocIndexer(" + self.name + ").__getitem__"
            return utils.fallback_call(
                self._unwrap, "__getitem__", key, __fireducks_reason=reason
            )

        # if not Series, self.obj must be a DataFrame
        if isinstance(key, tuple) and len(key) == 2:
            r, c = key
            # get single element
            if isinstance(r, int) and isinstance(c, int):
                if (
                    fireducks.core.get_fireducks_options().fast_fallback
                    and self.obj._fireducks_meta.is_cached()
                ):  # read from cache
                    cache = self.obj._fireducks_meta.get_cache()
                    return cache.iloc[key]
                else:
                    projected = _take_col(c)
                    value = ir.iloc_scalar(projected._value, r)
                    return _Scalar(value)._unwrap()

            is_supported_colp = isinstance(
                c, (int, range, slice, list, tuple, Series, Index, np.ndarray)
            )

            if is_supported_colp:
                # let's project target columns first
                if is_full_slice(c):  # e.g, iloc[:, :]
                    projected = self.obj
                elif isinstance(c, int):  # e.g, iloc[:, 0]
                    projected = _take_col(c)
                else:  # e.g., iloc[:, :2], iloc[:, [0,3]]
                    projected = self.obj._take(c, axis=1)

                # let's perform slice/filter rows on projected columns
                if is_full_slice(r):
                    # projected-columns, all-rows
                    return projected
                if isinstance(r, range):
                    # projected-columns, sliced-rows
                    r = slice(r.start, r.stop, r.step)
                    return projected._slice(r)
                elif isinstance(r, slice):
                    # projected-columns, sliced-rows
                    return projected._slice(r)
                elif isinstance(r, (list, Series, Index, np.ndarray)):
                    # projected-columns, filter-rows as per given row-indices
                    return projected._take(r, axis=0)

        reason = "fallback _ILocIndexer(" + self.name + ").__getitem__"
        return utils.fallback_call(
            self._unwrap, "__getitem__", key, __fireducks_reason=reason
        )
